from collections.abc import Iterable
from typing import Any, Generic, TypeVar

from typing_extensions import override

from bolinette.core import expressions
from bolinette.core.expressions.exceptions import ExpressionError
from bolinette.core.types import Type


class ExpressionNode:
    @override
    def __getattribute__(self, __name: str) -> Any:
        if __name == "__class__":
            return object.__getattribute__(self, "__class__")
        if __name == "__dict__":
            return object.__getattribute__(self, "__dict__")
        if __name == "__str__":
            return object.__getattribute__(self, "__str__")
        if __name == "__repr__":
            return object.__getattribute__(self, "__repr__")
        return AttributeNode(self, __name)

    def __getitem__(self, key: Any) -> "ElementNode[Any]":
        return ElementNode(self, key)

    def __expr_get_value__(self, obj: object) -> Any:
        raise ExpressionError(self, f"Cannot get value from {type(self).__name__}")

    def __expr_set_value__(self, obj: object, value: Any) -> None:
        raise ExpressionError(self, f"Cannot set value of {type(self).__name__}")

    def __expr_get_attribute__(self) -> str:
        raise ExpressionError(self, f"Cannot get attribute name of {type(self).__name__}")

    def __expr_format__(self, depth: int | None) -> str:
        raise ExpressionError(self, f"Cannot format {type(self).__name__}")

    def __expr_get_parents__(self) -> "Iterable[ExpressionNode]":
        raise ExpressionError(self, f"Cannot get parents from {type(self).__name__}")

    @override
    def __str__(self) -> str:
        return expressions.ExpressionTree.format(self)

    @override
    def __repr__(self) -> str:
        return f"<{type(self).__qualname__}: {expressions.ExpressionTree.format(self)}>"


class RootNode(ExpressionNode):
    def __init__(self, origin: Type[Any] | None = None) -> None:
        self.origin = origin

    @override
    def __expr_get_value__(self, obj: object) -> Any:
        return obj

    @override
    def __expr_format__(self, depth: int | None) -> str:
        if depth is not None and depth <= 0:
            return ""
        origin = object.__getattribute__(self, "origin")
        if origin is None:
            return "$"
        return str(origin)

    @override
    def __expr_get_parents__(self) -> Iterable[ExpressionNode]:
        return []


class ChildNode(ExpressionNode):
    def __init__(self, parent: ExpressionNode) -> None:
        self.parent = parent

    @override
    def __expr_get_parents__(self) -> Iterable[ExpressionNode]:
        parent = object.__getattribute__(self, "parent")
        return [parent]


class AttributeNode(ChildNode):
    def __init__(self, parent: ExpressionNode, attr: str) -> None:
        ChildNode.__init__(self, parent)
        self.attr = attr

    @override
    def __expr_get_value__(self, obj: object) -> Any:
        parent = object.__getattribute__(self, "parent")
        attr = object.__getattribute__(self, "attr")
        return getattr(expressions.ExpressionTree.get_value(parent, obj), attr)

    @override
    def __expr_set_value__(self, obj: object, value: Any) -> None:
        parent = object.__getattribute__(self, "parent")
        attr = object.__getattribute__(self, "attr")
        setattr(expressions.ExpressionTree.get_value(parent, obj), attr, value)

    @override
    def __expr_get_attribute__(self) -> str:
        return object.__getattribute__(self, "attr")

    @override
    def __expr_format__(self, depth: int | None) -> str:
        attr = object.__getattribute__(self, "attr")
        if depth is not None:
            if depth <= 1:
                return f"{attr}"
            depth -= 1
        parent = object.__getattribute__(self, "parent")
        return f"{expressions.ExpressionTree.format(parent, max_depth=depth)}.{attr}"


K = TypeVar("K")


class ElementNode(ChildNode, Generic[K]):
    def __init__(self, parent: ExpressionNode, key: K) -> None:
        ChildNode.__init__(self, parent)
        self.key = key

    @override
    def __expr_get_value__(self, obj: object) -> Any:
        parent = object.__getattribute__(self, "parent")
        key = object.__getattribute__(self, "key")
        return expressions.ExpressionTree.get_value(parent, obj)[key]

    @override
    def __expr_set_value__(self, obj: object, value: Any) -> None:
        parent = object.__getattribute__(self, "parent")
        key = object.__getattribute__(self, "key")
        expressions.ExpressionTree.get_value(parent, obj)[key] = value

    @override
    def __expr_format__(self, depth: int | None) -> str:
        key = object.__getattribute__(self, "key")
        if isinstance(key, str):
            key_s = f"'{key}'"
        else:
            key_s = str(key)
        if depth is not None:
            if depth <= 1:
                return f"[{key_s}]"
            depth -= 1
        parent = object.__getattribute__(self, "parent")
        return f"{expressions.ExpressionTree.format(parent, max_depth=depth)}[{key_s}]"



from typing import Any

from bolinette.core.expressions.exceptions import AttributeChainError, MaxDepthExpressionError
from bolinette.core.expressions.nodes import AttributeNode, ExpressionNode, RootNode
from bolinette.core.types import Type


class ExpressionTree:
    def __init__(self) -> None:
        raise TypeError(f"Cannot instantiate {ExpressionTree}")

    @staticmethod
    def new(origin: Type[Any] | None = None) -> RootNode:
        return RootNode(origin)

    @staticmethod
    def get_value(expr: ExpressionNode, obj: object) -> Any:
        return object.__getattribute__(expr, "__expr_get_value__")(obj)

    @staticmethod
    def set_value(expr: ExpressionNode, obj: object, value: Any) -> None:
        return object.__getattribute__(expr, "__expr_get_value__")(obj, value)

    @staticmethod
    def get_attribute(expr: ExpressionNode) -> str:
        return object.__getattribute__(expr, "__expr_get_attribute__")()

    @staticmethod
    def format(expr: ExpressionNode, *, max_depth: int | None = None) -> str:
        return object.__getattribute__(expr, "__expr_format__")(max_depth)

    @staticmethod
    def ensure_attribute_chain(
        expr: ExpressionNode,
        *,
        max_depth: int | None = None,
        origin: ExpressionNode | None = None,
    ) -> None:
        if max_depth is not None:
            if max_depth == 0:
                if object.__getattribute__(expr, "__class__") != RootNode:
                    raise MaxDepthExpressionError(origin or expr)
            elif object.__getattribute__(expr, "__class__") != AttributeNode:
                raise AttributeChainError(origin or expr)
        parents: list[ExpressionNode] = object.__getattribute__(expr, "__expr_get_parents__")()
        for parent in parents:
            ExpressionTree.ensure_attribute_chain(
                parent,
                max_depth=max_depth - 1 if max_depth is not None else None,
                origin=origin or expr,
            )



import os
from pathlib import Path
from typing import Any

import yaml

from bolinette.core import Cache, Logger, meta
from bolinette.core.environment.sections import EnvironmentSection, EnvSectionMeta
from bolinette.core.exceptions import EnvironmentError
from bolinette.core.expressions import ExpressionTree
from bolinette.core.injection import Injection, init_method
from bolinette.core.mapping import Mapper
from bolinette.core.types import Type


class Environment:
    _OS_ENV_PREFIX = "BLNT_"

    def __init__(self, cache: Cache, inject: Injection, logger: "Logger[Environment]") -> None:
        self._cache = cache
        self._inject = inject
        self._logger = logger
        self._env_folder = Path.cwd() / "env"
        self.profile: str
        self.config: dict[str, Any] = {}

    @init_method
    def init_profile(self) -> None:
        def read_profile() -> str:
            try:
                with open(self._env_folder / ".profile") as f:
                    for line in f:
                        return line.strip(" \n")
            except FileNotFoundError:
                self._logger.warning(".profile not found in env folder, defaulting to 'development'")
            return "development"

        self.profile = read_profile()

    @init_method
    def init(self) -> None:
        if EnvironmentSection in self._cache:
            for cls in self._cache.get(EnvironmentSection, hint=type):
                self._inject.add(cls, "singleton", before_init=[init_section])

        stack = [
            self._init_from_os(),
            self._init_from_file("env.yaml"),
            self._init_from_file(f"env.{self.profile}.yaml"),
            self._init_from_file(f"env.local.{self.profile}.yaml"),
        ]

        merged: dict[str, Any] = {}
        for node in stack:
            for name, section in node.items():
                if name not in merged:
                    merged[name] = {}
                for key, value in section.items():
                    merged[name][key] = value
        self.config = merged

    @staticmethod
    def _init_from_os() -> dict[str, dict[str, Any]]:
        _vars: dict[str, str] = {}
        prefix_len = len(Environment._OS_ENV_PREFIX)
        for var in os.environ:
            if var.startswith(Environment._OS_ENV_PREFIX):
                _vars[var[prefix_len:]] = os.environ[var]
        _env: dict[str, Any] = {}
        for key, value in _vars.items():
            path = [s.lower() for s in key.split("__")]
            _node = _env
            for p in path[:-1]:
                if p not in _node:
                    _node[p] = {}
                _node = _node[p]
            if not isinstance(_node, dict) or path[-1] in _node:
                raise EnvironmentError(
                    f"OS variable '{Environment._OS_ENV_PREFIX}{key}' conflicts with other variables"
                )
            _node[path[-1]] = value
        return _env

    def _init_from_file(self, file_name: str) -> dict[str, dict[str, Any]]:
        try:
            with open(self._env_folder / file_name) as f:
                return yaml.safe_load(f)
        except FileNotFoundError:
            return {}


def init_section(section: object, env: Environment, mapper: Mapper) -> None:
    section_name = meta.get(type(section), EnvSectionMeta).name
    if section_name not in env.config:
        raise EnvironmentError(f"No '{section_name}' section was found in the environment files")
    mapper.map(
        dict[str, Any],
        type(section),
        env.config[section_name],
        section,
        src_expr=ExpressionTree.new(Type(Environment))[section_name],
    )



import sys
from argparse import ArgumentParser
from collections.abc import Awaitable, Callable
from typing import Any, Protocol

from bolinette.core import Cache, Logger, meta
from bolinette.core.command.command import Argument, ArgumentMeta, CommandMeta
from bolinette.core.exceptions import InitError
from bolinette.core.injection import init_method


class _SubParsersAction(Protocol):
    def add_parser(self, name: str, *, help: str | None = None) -> ArgumentParser:
        ...


class Parser:
    def __init__(
        self,
        logger: Logger["Parser"],
        cache: Cache,
    ):
        self._cache = cache
        self._logger = logger
        self._factories = {
            "argument": self._create_argument,
            "option": self._create_option,
            "flag": self._create_flag,
        }
        self._functions: list[Callable[..., Awaitable[None]]] = []
        self._commands: dict[str, Callable[..., Awaitable[None]]] = {}
        self._sub_commands: dict[str, Any] = {}

    @init_method
    def init(self):
        if CommandMeta in self._cache:
            self._functions = self._cache.get(CommandMeta)

    def parse_command(
        self,
    ) -> tuple[Callable[..., Awaitable[int | None]], dict[str, Any]]:
        tree = self._parse_commands()
        parser = ArgumentParser(description="Bolinette Framework")
        sub_parsers = parser.add_subparsers()
        self._build_parsers(tree, sub_parsers, [])
        parsed = vars(parser.parse_args())
        if "__blnt_cmd__" in parsed:
            cmd = parsed.pop("__blnt_cmd__")
            if "__blnt_path__" in parsed:
                del parsed["__blnt_path__"]
            return (self._commands[cmd], parsed)
        elif "__blnt_path__" in parsed:
            self._logger.error(self._sub_commands[parsed["__blnt_path__"]].format_help())
            sys.exit(1)
        else:
            self._logger.error(parser.format_help())
            sys.exit(1)

    def _parse_commands(self) -> dict[str, Any]:
        command_tree: dict[str, Any] = {}
        for func in self._functions:
            _cmd = meta.get(func, CommandMeta)
            cur_node = command_tree
            path = _cmd.path.split(" ")
            for elem in path[:-1]:
                if elem not in cur_node:
                    cur_node[elem] = {}
                cur_node = cur_node[elem]
            elem = path[-1]
            if elem in cur_node:
                raise InitError(f"Conflict with '{_cmd.path}' command")
            cur_node[elem] = func
            self._commands[_cmd.path] = func
        return command_tree

    def _build_parsers(
        self,
        command_tree: dict[str, Any],
        sub_parsers: _SubParsersAction,
        path: list[str],
    ):
        for name, elem in command_tree.items():
            if meta.has(elem, CommandMeta):
                _cmd = meta.get(elem, CommandMeta)
                sub_parser = sub_parsers.add_parser(name, help=_cmd.summary)
                if meta.has(elem, ArgumentMeta):
                    for arg in meta.get(elem, ArgumentMeta):
                        self._factories[arg.arg_type](arg, sub_parser)
                sub_parser.set_defaults(__blnt_cmd__=_cmd.path)
            else:
                sub_parser = sub_parsers.add_parser(name, help=self._build_help(elem, [*path, name]))
                sub_parser.set_defaults(__blnt_path__=name)
                self._sub_commands[name] = sub_parser
                self._build_parsers(elem, sub_parser.add_subparsers(), [*path, name])

    @staticmethod
    def _build_help(command_tree: dict[str, Any], path: list[str]):
        commands = [f'"{" ".join([*path, x])}"' for x in command_tree]
        return "Sub-commands: " + ", ".join(commands)

    @staticmethod
    def _create_parser_arg(
        arg: Argument,
        *,
        optional: bool = False,
        use_flag: bool = False,
        action: str | None = None,
    ):
        args: list[Any] = []
        kwargs: dict[str, Any] = {}
        if optional:
            args.append(f"--{arg.name}")
        else:
            args.append(arg.name)
        if use_flag and arg.flag is not None:
            args.append(f"-{arg.flag}")
        if arg.summary is not None:
            kwargs["help"] = arg.summary
        if action is not None:
            kwargs["action"] = action
        if arg.value_type is not None:
            kwargs["type"] = arg.value_type
        return args, kwargs

    @staticmethod
    def _create_argument(arg: Argument, parser: ArgumentParser):
        args, kwargs = Parser._create_parser_arg(arg, optional=False, use_flag=False, action=None)
        parser.add_argument(*args, **kwargs)

    @staticmethod
    def _create_option(arg: Argument, parser: ArgumentParser):
        args, kwargs = Parser._create_parser_arg(arg, optional=True, use_flag=True, action=None)
        parser.add_argument(*args, **kwargs)

    @staticmethod
    def _create_flag(arg: Argument, parser: ArgumentParser):
        args, kwargs = Parser._create_parser_arg(arg, optional=True, use_flag=True, action="store_true")
        parser.add_argument(*args, **kwargs)



import json
from collections.abc import Iterable
from typing import Any

from typing_extensions import override


class JsonObjectEncoder(json.JSONEncoder):
    @override
    def default(self, o: Any) -> Any:
        return self.obj_to_primitives(o)

    def obj_to_primitives(self, o: object) -> Any:
        if isinstance(o, int | float | bool | str):
            return o
        if isinstance(o, Iterable):
            return list(*o)
        return {k: v for k, v in vars(o).items()}



import inspect
from collections.abc import Callable
from typing import Any, Generic, ParamSpec, TypeVar, get_type_hints

from typing_extensions import override

from bolinette.core.types import Type, TypeVarLookup

FuncP = ParamSpec("FuncP")
FuncT = TypeVar("FuncT")
InitP = ParamSpec("InitP")


class Function(Generic[FuncP, FuncT]):
    __slots__: list[str] = ["func", "bound_to"]

    def __init__(self, func: Callable[FuncP, FuncT]) -> None:
        self.func = func
        self.bound_to = getattr(self.func, "__self__", None)

    def parameters(self) -> dict[str, inspect.Parameter]:
        return {**inspect.signature(self.func).parameters}

    def annotations(self, *, lookup: TypeVarLookup[Any] | None = None) -> dict[str, Any]:
        return {
            n: self._transform_annotation(c, lookup) for n, c in get_type_hints(self.func, include_extras=True).items()
        }

    def __call__(self, *args: FuncP.args, **kwargs: FuncP.kwargs) -> FuncT:
        return self.func(*args, **kwargs)

    @staticmethod
    def _transform_annotation(anno: Any, lookup: TypeVarLookup[Any] | None) -> Any:
        if anno in (None, Ellipsis):
            return anno
        return Type(anno, lookup=lookup)

    @override
    def __str__(self) -> str:
        return f"{self.func.__qualname__}"

    @override
    def __repr__(self) -> str:
        return f"<Function {self!s}>"

    @override
    def __hash__(self) -> int:
        return hash(self.func)



import random
import re
import string
from collections.abc import Sequence
from typing import Any, Final


class StringUtils:
    __regex__: Final[dict[str, list[re.Pattern[str]]]] = {
        "snake": [
            re.compile(r"(.)([A-Z][a-z]+)"),
            re.compile(r"_+([A-Z])"),
            re.compile(r"([a-z0-9])([A-Z])"),
        ]
    }

    @staticmethod
    def to_snake_case(string: str):
        string = StringUtils.__regex__["snake"][0].sub(r"\1_\2", string)
        string = StringUtils.__regex__["snake"][1].sub(r"_\1", string)
        string = StringUtils.__regex__["snake"][2].sub(r"\1_\2", string)
        return string.lower()

    @staticmethod
    def capitalize(string: str):
        if not string:
            return string
        return f"{string[0].upper()}{string[1:]}"

    @staticmethod
    def random_string(length: int) -> str:
        return "".join(random.choices(string.ascii_letters + string.digits, k=length))

    @staticmethod
    def format_list(collection: Sequence[Any], *, sep: str = ", ", final_sep: str | None = None) -> str:
        formatted: list[str] = []
        cnt = len(collection)
        for i, e in enumerate(collection):
            formatted.append(str(e))
            if i != cnt - 1:
                if i == cnt - 2 and final_sep:
                    formatted.append(final_sep)
                else:
                    formatted.append(sep)
        return "".join(formatted)



from collections.abc import Callable
from typing import Any, Generic, Literal, TypeVar

from typing_extensions import override

from bolinette.core.exceptions import InjectionError
from bolinette.core.types import Type

InstanceT = TypeVar("InstanceT")


InjectionStrategy = Literal["singleton", "scoped", "transient", "immediate"]
AddStrategy = Literal["singleton", "scoped", "transient"]


class RegisteredType(Generic[InstanceT]):
    __slots__ = ("t", "strategy", "args", "named_args", "before_init", "after_init")

    def __init__(
        self,
        t: Type[InstanceT],
        strategy: InjectionStrategy,
        args: list[Any],
        named_args: dict[str, Any],
        before_init: list[Callable[..., None]],
        after_init: list[Callable[..., None]],
    ) -> None:
        self.t = t
        self.strategy = strategy
        self.args = args
        self.named_args = named_args
        self.before_init = before_init
        self.after_init = after_init

    @override
    def __repr__(self) -> str:
        return f"<RegisteredType {self.t}: {self.strategy}>"


class RegisteredTypeBag(Generic[InstanceT]):
    __slots__ = ("_cls", "_match_all", "_types")

    def __init__(self, cls: type[InstanceT]) -> None:
        self._cls = cls
        self._match_all: RegisteredType[InstanceT] | None = None
        self._types: dict[int, RegisteredType[InstanceT]] = {}

    def has_type(self, t: Type[InstanceT]) -> bool:
        return hash(t) in self._types

    def has_match_all(self) -> bool:
        return self._match_all is not None

    def is_registered(self, t: Type[InstanceT]) -> bool:
        return self.has_type(t) or self.has_match_all()

    def get_type(self, t: Type[Any]) -> RegisteredType[InstanceT]:
        if (h := hash(t)) in self._types:
            return self._types[h]
        if self._match_all is not None:
            return self._match_all
        raise InjectionError(f"Type {self._cls} has not been registered with parameters {t.vars}")

    def set_match_all(
        self,
        t: Type[InstanceT],
        strategy: InjectionStrategy,
        args: list[Any],
        named_args: dict[str, Any],
        before_init: list[Callable[[Any], None]],
        after_init: list[Callable[[Any], None]],
    ) -> RegisteredType[InstanceT]:
        r_type = RegisteredType(t, strategy, args, named_args, before_init, after_init)
        self._match_all = r_type
        return r_type

    def add_type(
        self,
        super_t: Type[InstanceT],
        t: Type[InstanceT],
        strategy: InjectionStrategy,
        args: list[Any],
        named_args: dict[str, Any],
        before_init: list[Callable[[Any], None]],
        after_init: list[Callable[[Any], None]],
    ) -> RegisteredType[InstanceT]:
        r_type = RegisteredType(t, strategy, args, named_args, before_init, after_init)
        self._types[hash(super_t)] = r_type
        return r_type

    @override
    def __repr__(self) -> str:
        return f"<RegisteredTypeBag {self._cls}: [{self._match_all}], [{len(self._types)}]>"



from typing import Any

from typing_extensions import override


class GenericMeta:
    def __init__(self, args: tuple[Any, ...]) -> None:
        self._args = args

    def __len__(self, /) -> int:
        return len(self._args)

    def __getitem__(self, index: int, /) -> Any:
        return self._args[index]

    @override
    def __hash__(self) -> int:
        return hash(self._args)

    @override
    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, GenericMeta):
            return self._args == value._args
        if isinstance(value, tuple):
            return self._args == value
        raise TypeError(type(value))



from typing import Any, TypeVar

from bolinette.core.types import Type

InstanceT = TypeVar("InstanceT")


class InjectionContext:
    __slots__ = ("_instances",)

    def __init__(self) -> None:
        self._instances: dict[int, Any] = {}

    def has_instance(self, t: Type[Any]) -> bool:
        return hash(t) in self._instances

    def set_instance(self, t: Type[InstanceT], instance: InstanceT) -> None:
        self._instances[hash(t)] = instance

    def get_instance(self, t: Type[InstanceT]) -> InstanceT:
        return self._instances[hash(t)]



from pathlib import Path

from setuptools import find_namespace_packages


def read_file(path: Path) -> str:
    with open(path) as f:
        return f.read()


def read_requirements(path: Path) -> list[str]:
    return [s for s in read_file(path).splitlines() if len(s)]


def project_packages(module: str) -> list[str]:
    return [m for m in find_namespace_packages() if m.startswith(module)]



from typing import Any, Generic, TypeVar

from typing_extensions import override

from bolinette.core import injection, meta
from bolinette.core.exceptions import InjectionError
from bolinette.core.injection.registration import RegisteredType
from bolinette.core.types import Type
from bolinette.core.utils import OrderedSet

InstanceT = TypeVar("InstanceT")


class InjectionHook(Generic[InstanceT]):
    __slots__ = ("t",)

    def __init__(self, t: Type[InstanceT]) -> None:
        self.t = t

    @override
    def __getattribute__(self, __name: str) -> Any:
        if __name in ("t", "__class__"):
            return object.__getattribute__(self, __name)
        raise InjectionError(
            f"Tried accessing member '{__name}' of an injected instance inside the __init__ method. "
            "Use @init_method to process logic at instantiation."
        )

    def __get__(self, *_) -> InstanceT:
        return None  # type: ignore


class InjectionProxy:
    __slots__ = ("name", "r_type", "t")

    def __init__(
        self,
        name: str,
        r_type: RegisteredType[InstanceT],
        t: Type[InstanceT],
    ) -> None:
        self.name = name
        self.r_type = r_type
        self.t = t

    def __get__(self, instance: Any, _) -> Any:
        inject = meta.get(instance, injection.Injection)
        if inject.__has_instance__(self.r_type):
            obj = inject.__get_instance__(self.r_type)
        else:
            obj = inject.__instantiate__(self.r_type, self.t, OrderedSet())
        setattr(instance, self.name, obj)
        return obj



from bolinette.core.utils.attrs import AttributeUtils as AttributeUtils
from bolinette.core.utils.strings import StringUtils as StringUtils
from bolinette.core.utils.collections import OrderedSet as OrderedSet



from typing import Any, TypeVar

from bolinette.core.exceptions import InternalError

T = TypeVar("T")


class BolinetteMetadata:
    def __init__(self) -> None:
        self._data: dict[type[Any], Any] = {}

    def __contains__(self, key: type[Any]) -> bool:
        return key in self._data

    def __getitem__(self, key: type[T]) -> T:
        if key not in self:
            raise KeyError(key)
        return self._data[key]

    def __setitem__(self, key: type[T], value: T) -> None:
        if not isinstance(value, key):
            raise TypeError(f"Type mismatch between {key} and {value}")
        self._data[key] = value


def get_meta_container(obj: Any) -> BolinetteMetadata:
    if "__blnt_meta__" not in vars(obj):
        _meta = BolinetteMetadata()
        obj.__blnt_meta__ = _meta
    else:
        _meta = obj.__blnt_meta__
        if not isinstance(_meta, BolinetteMetadata):
            raise InternalError(
                f"Metadata container in {obj} has been overwritten. "
                "Please do not use '__blnt_meta__' as an attribute in any class"
            )
    return _meta


class _MetaFunctions:
    @staticmethod
    def has(obj: Any, cls: type[Any], /) -> bool:
        if not hasattr(obj, "__dict__"):
            return False
        container = get_meta_container(obj)
        return cls in container

    @staticmethod
    def set(obj: Any, _meta: T, /, *, cls: type[T] | None = None) -> None:
        if cls is None:
            cls = type(_meta)
        else:
            if not isinstance(_meta, cls):
                raise TypeError(f"Type mismatch between {cls} and {_meta}")
        container = get_meta_container(obj)
        container[cls] = _meta

    @staticmethod
    def get(obj: Any, cls: type[T], /, *, default: T | None = None) -> T:
        container = get_meta_container(obj)
        if cls not in container and default is not None:
            return default
        return container[cls]


meta = _MetaFunctions()



import importlib
import pkgutil
import sys
from collections.abc import Callable
from types import ModuleType
from typing import NoReturn

import bolinette
from bolinette.core import Extension, Logger, __user_cache__, meta
from bolinette.core.cache import Cache
from bolinette.core.command import Parser
from bolinette.core.environment import Environment
from bolinette.core.exceptions import InitError
from bolinette.core.injection import Injection, require


class Bolinette:
    def __init__(
        self,
        *,
        profile: str | None = None,
    ) -> None:
        self.cache = Cache()

        self.extensions = self._load_extensions()
        for ext in self.extensions:
            ext.add_cached(self.cache)

        self.cache |= __user_cache__

        self._inject = Injection(self.cache)
        meta.set(self, self._inject)
        self._env = self._inject.require(Environment)
        self.cache.debug = self._env.config["core"]["debug"]

        self._inject.add(Bolinette, "singleton", instance=self)
        self._inject.__hook_proxies__(self)

        self._logger = self._inject.require(Logger[Bolinette])
        self._logger.info(f"Loaded Bolinette with extensions: {', '.join(e.name for e in self.extensions)}")

    @property
    def injection(self) -> Injection:
        return self._inject

    def _load_extensions(self) -> list[Extension]:
        def iter_namespace(module: ModuleType) -> list[str]:
            return [info[1] for info in pkgutil.iter_modules(module.__path__, module.__name__ + ".")]

        blnt_modules = [importlib.import_module(name) for name in iter_namespace(bolinette)]
        extensions: list[Extension] = []
        for module in blnt_modules:
            if not hasattr(module, "__blnt_ext__") or not isinstance(ext := module.__blnt_ext__, Extension):
                raise InitError(f"{module.__name__} is not a valid Bolinette extension module")
            extensions.append(ext)
        return Extension.sort_extensions(extensions)

    def _set_default_profile(self) -> str:
        self._logger.warning(
            "No profile set, defaulting to 'development'.",
            "Be sure to set the current profile in a .profile file in the env folder",
        )
        return "development"

    async def startup(self) -> None:
        pass

    @require(Parser)
    def _parser(self):
        ...

    async def run_from_args(self) -> NoReturn:
        cmd, args = self._parser.parse_command()
        result = await self._inject.call(cmd, named_args=args)
        sys.exit(result or 0)


def main_func(func: Callable[[], Bolinette]) -> Callable[[], Bolinette]:
    setattr(func, "__blnt__", "__blnt_main__")
    return func



from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, Concatenate, Literal, ParamSpec, Protocol, TypeVar, overload

from typing_extensions import override

from bolinette.core import Cache, GenericMeta, __user_cache__, meta
from bolinette.core.exceptions import InjectionError
from bolinette.core.injection.context import InjectionContext
from bolinette.core.injection.decorators import InitMethodMeta, InjectionParamsMeta, InjectionSymbol
from bolinette.core.injection.hook import InjectionHook, InjectionProxy
from bolinette.core.injection.registration import AddStrategy, InjectionStrategy, RegisteredType, RegisteredTypeBag
from bolinette.core.injection.resolver import ArgResolverMeta, ArgResolverOptions, ArgumentResolver, DefaultArgResolver
from bolinette.core.types import Function, Type, TypeVarLookup
from bolinette.core.utils import OrderedSet
from bolinette.core.utils.strings import StringUtils

FuncP = ParamSpec("FuncP")
FuncT = TypeVar("FuncT")
InstanceT = TypeVar("InstanceT")


class Injection:
    __slots__: list[str] = ["cache", "_global_ctx", "_types", "_default_resolver", "_arg_resolvers", "_callbacks"]
    _ADD_INSTANCE_STRATEGIES = ("singleton",)
    _REQUIREABLE_STRATEGIES = ("singleton", "transient")

    def __init__(
        self,
        cache: Cache,
        global_ctx: InjectionContext | None = None,
        types: "dict[type[Any], RegisteredTypeBag[Any]] | None" = None,
    ) -> None:
        self.cache = cache
        self._callbacks: Iterable[InjectionCallback] = []
        self._global_ctx = global_ctx or InjectionContext()
        self._types = types if types is not None else self._pickup_types(cache)
        self._arg_resolvers: list[ArgumentResolver] = []
        self._default_resolver: ArgumentResolver = DefaultArgResolver()
        self._add_type_instance(Type(Cache), Type(Cache), False, "singleton", [], {}, [], [], cache, safe=True)
        self._callbacks = self._pickup_callbacks()
        self._add_type_instance(Type(Injection), Type(Injection), False, "singleton", [], {}, [], [], self, safe=True)
        self._arg_resolvers = self._pickup_resolvers(cache)

    @property
    def registered_types(self):
        return dict(self._types)

    @staticmethod
    def _pickup_types(cache: Cache) -> "dict[type[Any], RegisteredTypeBag[Any]]":
        if InjectionSymbol not in cache:
            return {}
        types: dict[type[Any], RegisteredTypeBag[Any]] = {}
        for cls in cache.get(InjectionSymbol, hint=type):
            t = Type(cls)
            if t.cls not in types:
                types[t.cls] = RegisteredTypeBag(t.cls)
            type_bag = types[t.cls]
            _meta = meta.get(t.cls, InjectionParamsMeta)
            if _meta.match_all:
                type_bag.set_match_all(
                    t,
                    _meta.strategy,  # type: ignore
                    _meta.args,
                    _meta.named_args,
                    _meta.before_init,
                    _meta.after_init,
                )
            else:
                type_bag.add_type(
                    t,
                    t,
                    _meta.strategy,  # type: ignore
                    _meta.args,
                    _meta.named_args,
                    _meta.before_init,
                    _meta.after_init,
                )
        return types

    def _pickup_resolvers(self, cache: Cache) -> "list[ArgumentResolver]":
        resolver_scoped: dict[type, bool] = {}
        resolver_priority: dict[type, int] = {}
        resolver_types: list[type[ArgumentResolver]] = []

        for t in cache.get(ArgumentResolver, hint=type[ArgumentResolver], raises=False):
            _meta = meta.get(t, ArgResolverMeta)
            resolver_priority[t] = _meta.priority
            resolver_scoped[t] = _meta.scoped
            resolver_types.append(t)
        resolver_types = sorted(resolver_types, key=lambda t: resolver_priority[t])

        resolvers: list[ArgumentResolver] = []
        for cls in resolver_types:
            if resolver_scoped[cls]:
                continue
            resolvers.append(self.instantiate(cls))

        return [*resolvers]

    def _pickup_callbacks(self) -> "Iterable[InjectionCallback]":
        return [
            self.instantiate(t) for t in self.cache.get(InjectionCallback, hint=type[InjectionCallback], raises=False)
        ]

    def __has_instance__(self, r_type: "RegisteredType[Any]") -> bool:
        return r_type.strategy == "singleton" and self._global_ctx.has_instance(r_type.t)

    def __get_instance__(self, r_type: "RegisteredType[InstanceT]") -> InstanceT:
        return self._global_ctx.get_instance(r_type.t)

    def __set_instance__(self, r_type: "RegisteredType[InstanceT]", instance: InstanceT) -> None:
        if r_type.strategy == "singleton":
            self._global_ctx.set_instance(r_type.t, instance)

    def _resolve_args(
        self,
        obj: Type[Any] | Function[..., Any],
        type_vars: tuple[Any, ...] | None,
        strategy: InjectionStrategy,
        vars_lookup: TypeVarLookup[Any] | None,
        immediate: bool,
        circular_guard: OrderedSet[Any],
        args: list[Any],
        named_args: dict[str, Any],
        additional_resolvers: list[ArgumentResolver],
    ) -> dict[str, Any]:
        if obj in circular_guard:
            call_chain = " -> ".join(str(f) for f in circular_guard) + f" -> {obj}"
            raise InjectionError(f"A circular call has been detected: {call_chain}", cls=circular_guard[0])
        circular_guard.add(obj)

        try:
            if isinstance(obj, Function):
                annotations = obj.annotations(lookup=vars_lookup)
            else:
                annotations = obj.init.annotations(lookup=vars_lookup)
        except NameError as exp:
            raise InjectionError(f"Type hint '{exp.name}' could not be resolved", func=obj) from exp
        func_params = obj.parameters()

        if any((n, p) for n, p in func_params.items() if p.kind in (p.POSITIONAL_ONLY, p.VAR_POSITIONAL)):
            raise InjectionError(
                "Positional only parameters and positional wildcards are not allowed",
                func=obj,
            )

        f_args: dict[str, Any] = {}
        _args = [*args]
        _named_args = {**named_args}

        for p_name, param in func_params.items():
            if param.kind == param.VAR_KEYWORD:
                for kw_name, kw_value in _named_args.items():
                    f_args[kw_name] = kw_value
                _named_args = {}
                break

            if _args:
                f_args[p_name] = _args.pop(0)
                continue
            if p_name in _named_args:
                f_args[p_name] = _named_args.pop(p_name)
                continue

            default_set = False
            default = None
            if param.default is not param.empty:
                default_set = True
                default = param.default

            if p_name not in annotations:
                if default_set:
                    f_args[p_name] = default
                    continue
                raise InjectionError("Annotation is required", func=obj, param=p_name)

            hint: Type[Any] = annotations[p_name]

            if hint.is_union:
                raise InjectionError("Type unions are not allowed", func=obj, param=p_name)

            for resolver in [*additional_resolvers, *self._arg_resolvers, self._default_resolver]:
                options = ArgResolverOptions(
                    self,
                    obj,
                    type_vars,
                    strategy,
                    p_name,
                    hint,
                    default_set,
                    default,
                    immediate,
                    circular_guard,
                )
                if resolver.supports(options):
                    arg_name, arg_value = resolver.resolve(options)
                    f_args[arg_name] = arg_value
                    break

        if _args or _named_args:
            raise InjectionError(
                f"Expected {len(func_params)} arguments, {len(args) + len(named_args)} given",
                func=obj,
            )

        return f_args

    def __hook_proxies__(self, instance: object) -> None:
        hooks: dict[str, Type[Any]] = {}
        cls = type(instance)
        cls_attrs: dict[str, Any] = dict(vars(cls))
        attr: InjectionHook[Any] | Any
        for name, attr in cls_attrs.items():
            if isinstance(attr, InjectionHook):
                delattr(cls, name)
                hooks[name] = attr.t
        instance_attrs: dict[str, Any] = dict(vars(instance))
        for name, attr in instance_attrs.items():
            if isinstance(attr, InjectionHook):
                delattr(instance, name)
                hooks[name] = attr.t
        for name, t in hooks.items():
            r_type = self._types[t.cls].get_type(t)
            setattr(cls, name, InjectionProxy(name, r_type, t))

    def _run_init_recursive(
        self,
        cls: type[InstanceT],
        instance: InstanceT,
        vars_lookup: TypeVarLookup[InstanceT] | None,
        circular_guard: OrderedSet[Any] | None,
    ) -> None:
        for base in cls.__bases__:
            self._run_init_recursive(base, instance, vars_lookup, circular_guard)
        for _, attr in vars(cls).items():
            if meta.has(attr, InitMethodMeta):
                self.call(attr, args=[instance], vars_lookup=vars_lookup, circular_guard=circular_guard)

    def _run_init_methods(
        self,
        r_type: "RegisteredType[InstanceT]",
        instance: InstanceT,
        vars_lookup: TypeVarLookup[InstanceT] | None,
        circular_guard: OrderedSet[Any],
    ):
        for method in r_type.before_init:
            self.call(method, args=[instance], circular_guard=circular_guard)
        self._run_init_recursive(r_type.t.cls, instance, vars_lookup, circular_guard)
        for method in r_type.after_init:
            self.call(method, args=[instance], circular_guard=circular_guard)

    def __instantiate__(
        self,
        r_type: "RegisteredType[InstanceT]",
        t: Type[InstanceT],
        circular_guard: OrderedSet[Any],
    ) -> InstanceT:
        vars_lookup = TypeVarLookup(t)
        func_args = self._resolve_args(
            r_type.t,
            t.vars,
            r_type.strategy,  # type: ignore
            vars_lookup,
            False,
            circular_guard,
            r_type.args,
            r_type.named_args,
            [],
        )
        instance = r_type.t.cls(**func_args)
        self.__hook_proxies__(instance)
        meta.set(instance, self, cls=Injection)
        meta.set(instance, GenericMeta(t.vars))
        self._run_init_methods(r_type, instance, vars_lookup, circular_guard)
        self.__set_instance__(r_type, instance)
        for callback in self._callbacks:
            callback(
                "instantiated",
                t,
                r_type.strategy,  # type: ignore
                instance,
            )
        return instance

    def is_registered(self, cls: type[Any] | Type[Any]) -> bool:
        if not isinstance(cls, Type):
            t = Type(cls)
        else:
            t = cls
        return t.cls in self._types and self._types[t.cls].is_registered(t)

    def call(
        self,
        func: Callable[..., FuncT],
        *,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        vars_lookup: TypeVarLookup[Any] | None = None,
        additional_resolvers: list[ArgumentResolver] | None = None,
        circular_guard: OrderedSet[Any] | None = None,
    ) -> FuncT:
        func_args = self._resolve_args(
            Function(func),
            None,
            "singleton",
            vars_lookup,
            True,
            circular_guard or OrderedSet(),
            args or [],
            named_args or {},
            additional_resolvers or [],
        )
        return func(**func_args)

    def instantiate(
        self,
        cls: type[InstanceT],
        *,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        additional_resolvers: list[ArgumentResolver] | None = None,
    ) -> InstanceT:
        t = Type(cls)
        init_args = self._resolve_args(
            t,
            t.vars,
            "immediate",
            None,
            True,
            OrderedSet(),
            args or [],
            named_args or {},
            additional_resolvers or [],
        )
        instance = cls(**init_args)
        self.__hook_proxies__(instance)
        meta.set(instance, self, cls=Injection)
        meta.set(instance, GenericMeta(t.vars))
        self._run_init_recursive(cls, instance, None, None)
        return instance

    def _add_type_instance(
        self,
        super_t: Type[InstanceT],
        t: Type[InstanceT],
        match_all: bool,
        strategy: InjectionStrategy,
        args: list[Any],
        named_args: dict[str, Any],
        before_init: list[Callable[[Any], None]],
        after_init: list[Callable[[Any], None]],
        instance: InstanceT | None,
        *,
        safe: bool = False,
    ) -> None:
        if super_t.cls not in self._types:
            self._types[super_t.cls] = RegisteredTypeBag(super_t)
        type_bag = self._types[super_t.cls]
        r_type: RegisteredType[InstanceT] | None = None
        if match_all:
            if not safe or not type_bag.has_match_all():
                r_type = type_bag.set_match_all(t, strategy, args, named_args, before_init, after_init)
        else:
            if not safe or not type_bag.has_type(super_t):
                r_type = type_bag.add_type(super_t, t, strategy, args, named_args, before_init, after_init)
        if instance is not None:
            if r_type is None:
                r_type = self._types[t.cls].get_type(t)
            if not safe or not self.__has_instance__(r_type):
                self.__set_instance__(r_type, instance)

    @overload
    def add(
        self,
        cls: type[InstanceT],
        strategy: AddStrategy,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        instance: InstanceT | None = None,
        before_init: list[Callable[Concatenate[InstanceT, FuncP], None]] | None = None,
        after_init: list[Callable[Concatenate[InstanceT, FuncP], None]] | None = None,
        match_all: bool = False,
        super_cls: type[InstanceT] | None = None,
        *,
        instantiate: Literal[True],
    ) -> InstanceT:
        pass

    @overload
    def add(
        self,
        cls: type[InstanceT],
        strategy: AddStrategy,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        instance: InstanceT | None = None,
        before_init: list[Callable[Concatenate[InstanceT, FuncP], None]] | None = None,
        after_init: list[Callable[Concatenate[InstanceT, FuncP], None]] | None = None,
        match_all: bool = False,
        super_cls: type[InstanceT] | None = None,
    ) -> None:
        pass

    def add(
        self,
        cls: type[InstanceT],
        strategy: AddStrategy,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        instance: InstanceT | None = None,
        before_init: list[Callable[Concatenate[InstanceT, FuncP], None]] | None = None,
        after_init: list[Callable[Concatenate[InstanceT, FuncP], None]] | None = None,
        match_all: bool = False,
        super_cls: type[InstanceT] | None = None,
        *,
        instantiate: bool = False,
    ) -> InstanceT | None:
        if super_cls is None:
            super_cls = cls
        t = Type(cls)
        if hasattr(t.cls, "__parameters__"):
            if TYPE_CHECKING:
                assert isinstance(t.cls, _GenericOrigin)
            if len(t.cls.__parameters__) != len(t.vars) and not match_all:
                raise InjectionError(
                    f"Type {t} requires {len(t.cls.__parameters__)} generic parameters and {len(t.vars)} were given"
                )
        super_t: Type[InstanceT] = Type(super_cls)
        if not issubclass(t.cls, super_t.cls):
            raise InjectionError(f"Type {t} does not inherit from type {super_t}")
        if instance is not None:
            if instantiate:
                raise InjectionError(
                    f"Cannot instantiate {t.cls} if an instance is provided",
                )
            if not isinstance(instance, t.cls):
                raise InjectionError(f"Object provided must an instance of type {t.cls}")
            if strategy not in self._ADD_INSTANCE_STRATEGIES:
                formatted_strategies = StringUtils.format_list(self._ADD_INSTANCE_STRATEGIES, final_sep=" or ")
                raise InjectionError(
                    f"Injection strategy for {t.cls} must be {formatted_strategies} if an instance is provided"
                )
        self._add_type_instance(
            super_t,
            t,
            match_all,
            strategy,
            args or [],
            named_args or {},
            before_init or [],
            after_init or [],
            instance,
        )
        if instantiate:
            return self.require(t.cls)  # type: ignore
        return None

    def require(self, cls: type[InstanceT]) -> InstanceT:
        t = Type(cls)
        if not self.is_registered(t):
            raise InjectionError(f"Type {t} is not a registered type in the injection system")
        r_type = self._types[t.cls].get_type(t)
        if r_type.strategy not in self._REQUIREABLE_STRATEGIES:
            formatted_strategies = StringUtils.format_list(self._REQUIREABLE_STRATEGIES, final_sep=" or ")
            raise InjectionError(
                f"Injection strategy for {t} must be {formatted_strategies} to be required in this context"
            )
        if self.__has_instance__(r_type):
            return self.__get_instance__(r_type)
        return self.__instantiate__(r_type, t, OrderedSet())

    def get_scoped_session(self) -> "ScopedInjection":
        return ScopedInjection(self.cache, self._global_ctx, InjectionContext(), self._types)


class ScopedInjection(Injection):
    __slots__: list[str] = ["_scoped_ctx"]

    _ADD_INSTANCE_STRATEGIES = ("singleton", "scoped")
    _REQUIREABLE_STRATEGIES = ("singleton", "scoped", "transient")

    def __init__(
        self,
        cache: Cache,
        global_ctx: InjectionContext,
        scoped_ctx: InjectionContext,
        types: "dict[type[Any], RegisteredTypeBag[Any]]",
    ) -> None:
        self._scoped_ctx = scoped_ctx
        super().__init__(cache, global_ctx, types)
        self._scoped_ctx.set_instance(Type(Injection), self)

    @override
    def __has_instance__(self, r_type: "RegisteredType[Any]") -> bool:
        return (r_type.strategy == "scoped" and self._scoped_ctx.has_instance(r_type.t)) or (
            r_type.strategy == "singleton" and self._global_ctx.has_instance(r_type.t)
        )

    @override
    def __get_instance__(self, r_type: "RegisteredType[InstanceT]") -> InstanceT:
        if self._scoped_ctx.has_instance(r_type.t):
            return self._scoped_ctx.get_instance(r_type.t)
        return self._global_ctx.get_instance(r_type.t)

    @override
    def __set_instance__(self, r_type: "RegisteredType[InstanceT]", instance: InstanceT) -> None:
        strategy = r_type.strategy
        if strategy == "scoped":
            self._scoped_ctx.set_instance(r_type.t, instance)
        if strategy == "singleton":
            self._global_ctx.set_instance(r_type.t, instance)

    @override
    def _pickup_resolvers(self, cache: Cache) -> "list[ArgumentResolver]":
        resolver_priority: dict[type, int] = {}
        resolver_types: list[type[ArgumentResolver]] = []

        for t in cache.get(ArgumentResolver, hint=type[ArgumentResolver], raises=False):
            _meta = meta.get(t, ArgResolverMeta)
            resolver_priority[t] = _meta.priority
            resolver_types.append(t)
        resolver_types = sorted(resolver_types, key=lambda t: resolver_priority[t])

        resolvers: list[ArgumentResolver] = []
        for cls in resolver_types:
            resolvers.append(self.instantiate(cls))

        return [*resolvers]

    @override
    def call(
        self,
        func: Callable[..., FuncT],
        *,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        vars_lookup: TypeVarLookup[Any] | None = None,
        additional_resolvers: list[ArgumentResolver] | None = None,
        circular_guard: OrderedSet[Any] | None = None,
    ) -> FuncT:
        func_args = self._resolve_args(
            Function(func),
            None,
            "scoped",
            vars_lookup,
            True,
            circular_guard or OrderedSet(),
            args or [],
            named_args or {},
            additional_resolvers or [],
        )
        return func(**func_args)


InjectionEvent = Literal["instantiated"]


class InjectionCallback(Protocol):
    def __call__(self, event: InjectionEvent, type: Type[Any], strategy: InjectionStrategy, instance: Any) -> None:
        ...


CallbackT = TypeVar("CallbackT", bound=InjectionCallback)


def injection_callback(*, cache: Cache) -> Callable[[type[CallbackT]], type[CallbackT]]:
    def decorator(callback: type[CallbackT]) -> type[CallbackT]:
        (cache or __user_cache__).add(InjectionCallback, callback)
        return callback

    return decorator


class _GenericOrigin(Protocol):
    __parameters__: tuple[TypeVar, ...]



from collections.abc import Sequence
from typing import Any

from bolinette.core.exceptions import BolinetteError, ParameterError
from bolinette.core.expressions import ExpressionNode
from bolinette.core.types import Type


class MappingError(BolinetteError, ParameterError):
    def __init__(self, message: str, *, src: ExpressionNode | None = None, dest: ExpressionNode | None = None) -> None:
        ParameterError.__init__(self, dest="Destination path '{}'", src="From source path '{}'")
        BolinetteError.__init__(self, self._format_params(message, dest=dest, src=src))
        self.src = src
        self.dest = dest


class SourceNotFoundError(MappingError):
    src: ExpressionNode
    dest: ExpressionNode

    def __init__(self, src: ExpressionNode, dest: ExpressionNode, t: Type[Any]) -> None:
        super().__init__(
            f"Source path not found, could not bind a None value to non nullable type {t}", src=src, dest=dest
        )


class DestinationNotNullableError(MappingError):
    src: ExpressionNode
    dest: ExpressionNode

    def __init__(self, src: ExpressionNode, dest: ExpressionNode, t: Type[Any]) -> None:
        super().__init__(f"Could not bind a None value to non nullable type {t}", src=src, dest=dest)


class InstantiationError(MappingError):
    dest: ExpressionNode

    def __init__(self, dest: ExpressionNode, t: Type[Any]) -> None:
        super().__init__(
            f"Could not instantiate type {t}, make sure the __init__ has no required parameters", dest=dest
        )


class IgnoreImpossibleError(MappingError):
    dest: ExpressionNode

    def __init__(self, dest: ExpressionNode, t: Type[Any]) -> None:
        super().__init__(f"Could not ignore attribute, type {t} is not nullable", dest=dest)


class UnionNotAllowedError(MappingError):
    def __init__(self, dest: ExpressionNode, t: Type[Any]) -> None:
        super().__init__(f"Destination type {t} is a union, please use use_type(...) in profile", dest=dest)


class TypeMismatchError(MappingError):
    src: ExpressionNode
    dest: ExpressionNode

    def __init__(self, src: ExpressionNode, dest: ExpressionNode, source: Type[Any], target: Type[Any]) -> None:
        super().__init__(f"Selected type {source} is not assignable to {target}", src=src, dest=dest)


class TypeNotIterableError(MappingError):
    src: ExpressionNode
    dest: ExpressionNode

    def __init__(self, src: ExpressionNode, dest: ExpressionNode, source: Type[Any], target: Type[Any]) -> None:
        super().__init__(f"Could not map non iterable type {source} to {target}", src=src, dest=dest)


class ImmutableCollectionError(MappingError):
    dest: ExpressionNode

    def __init__(self, dest: ExpressionNode) -> None:
        super().__init__("Could not use an existing tuple instance, tuples are immutable", dest=dest)


class ConvertionError(MappingError):
    src: ExpressionNode
    dest: ExpressionNode

    def __init__(self, src: ExpressionNode, dest: ExpressionNode, value: Any, target: Type[Any]) -> None:
        super().__init__(f"Could not convert value '{value}' to {target}", src=src, dest=dest)
        self.target = target


class ValidationError(MappingError):
    def __init__(self, errors: Sequence[MappingError]) -> None:
        super().__init__("Mapping errors raised during validation")
        self.errors = errors



from collections.abc import Callable, Iterable
from typing import Any, Protocol, TypeGuard, TypeVar, overload

from typing_extensions import override

from bolinette.core import Cache, __user_cache__, meta
from bolinette.core.expressions import ExpressionNode, ExpressionTree
from bolinette.core.injection import Injection, init_method
from bolinette.core.mapping.exceptions import (
    ConvertionError,
    DestinationNotNullableError,
    IgnoreImpossibleError,
    ImmutableCollectionError,
    InstantiationError,
    MappingError,
    SourceNotFoundError,
    TypeMismatchError,
    TypeNotIterableError,
    UnionNotAllowedError,
    ValidationError,
)
from bolinette.core.mapping.profiles import Profile
from bolinette.core.mapping.sequence import IgnoreAttribute, MapFromAttribute, MappingSequence
from bolinette.core.types import Type


class NoInitDestination(Protocol):
    def __init__(self) -> None:
        pass


SrcT = TypeVar("SrcT", bound=object)
DestT = TypeVar("DestT", bound=object)
NoInitDestT = TypeVar("NoInitDestT", bound=NoInitDestination)
TargetT = TypeVar("TargetT")


class Mapper:
    def __init__(self) -> None:
        self._sequences: dict[int, MappingSequence[Any, Any]] = {}
        self._type_mappers: dict[Type[Any], type[TypeMapper[Any]]] = {}
        self._default_mapper: type[TypeMapper[object]] = DefaultTypeMapper

    @init_method
    def _init_profiles(self, cache: Cache, inject: Injection) -> None:
        completed: dict[int, MappingSequence[Any, Any]] = {}
        for cls in cache.get(Profile, hint=type[Profile], raises=False):
            profile = inject.instantiate(cls)
            for sequence in profile.sequences:
                sequence.complete(completed)
                completed[hash(sequence)] = sequence
        self._sequences = completed

    @init_method
    def _init_type_mappers(self, cache: Cache) -> None:
        mappers: dict[Type[Any], type[TypeMapper[Any]]] = {}
        for cls in cache.get(TypeMapperMeta, hint=type[TypeMapper[Any]], raises=False):
            _m = meta.get(cls, TypeMapperMeta)
            mappers[_m.t] = cls
        self._type_mappers = mappers

    def add_type_mapper(self, t: Type[TargetT], mapper: "type[TypeMapper[TargetT]]") -> None:
        self._type_mappers[t] = mapper

    def set_default_type_mapper(self, mapper: "type[TypeMapper[object]]") -> None:
        self._default_mapper = mapper

    @overload
    def map(
        self,
        src_cls: type[SrcT],
        dest_cls: type[NoInitDestT],
        src: SrcT,
        *,
        src_expr: ExpressionNode | None = None,
        dest_expr: ExpressionNode | None = None,
        validate: bool = False,
    ) -> NoInitDestT:
        pass

    @overload
    def map(
        self,
        src_cls: type[SrcT],
        dest_cls: type[DestT],
        src: SrcT,
        dest: DestT,
        *,
        src_expr: ExpressionNode | None = None,
        dest_expr: ExpressionNode | None = None,
        validate: bool = False,
    ) -> DestT:
        pass

    def map(
        self,
        src_cls: type[SrcT],
        dest_cls: type[DestT],
        src: SrcT,
        dest: DestT | None = None,
        *,
        src_expr: ExpressionNode | None = None,
        dest_expr: ExpressionNode | None = None,
        validate: bool = False,
    ) -> DestT:
        src_t = Type(src_cls)
        dest_t = Type(dest_cls)

        exc_grp: list[MappingError] | None
        if validate:
            exc_grp = []
        else:
            exc_grp = None

        mapped = MappingRunner(self._sequences, self._type_mappers, self._default_mapper).map(
            src_expr or ExpressionTree.new(src_t),
            src_t,
            dest_expr or ExpressionTree.new(dest_t),
            dest_t,
            src,
            dest,
            exc_grp,
        )
        if exc_grp is not None and len(exc_grp) > 0:
            raise ValidationError(exc_grp)
        return mapped


class MappingRunner:
    __slots__ = ("sequences", "mappers", "default_mapper")

    def __init__(
        self,
        sequences: dict[int, MappingSequence[Any, Any]],
        mappers: "dict[Type[Any], type[TypeMapper[Any]]]",
        default_mapper: "type[TypeMapper[object]]",
    ) -> None:
        self.sequences = sequences
        self.mappers = mappers
        self.default_mapper = default_mapper

    def map(
        self,
        src_expr: ExpressionNode,
        src_t: Type[SrcT],
        dest_expr: ExpressionNode,
        dest_t: Type[DestT],
        src: SrcT,
        dest: DestT | None,
        exc_grp: list[MappingError] | None,
    ) -> DestT:
        if src is None:
            if not dest_t.nullable:
                exc = DestinationNotNullableError(src_expr, dest_expr, dest_t)
                if exc_grp is None:
                    raise exc
                exc_grp.append(exc)
                return None  # type: ignore
            return None  # type: ignore
        mapper_cls: type[TypeMapper[Any]] = self.mappers[dest_t] if dest_t in self.mappers else self.default_mapper
        mapper = mapper_cls(self)
        return mapper.map(src_expr, src_t, dest_expr, dest_t, src, dest, exc_grp)


class TypeMapper(Protocol[TargetT]):
    def __init__(self, runner: MappingRunner) -> None:
        pass

    def map(
        self,
        src_expr: ExpressionNode,
        src_t: Type[Any],
        dest_expr: ExpressionNode,
        dest_t: Type[TargetT],
        src: Any,
        dest: TargetT | None,
        exc_grp: list[MappingError] | None,
    ) -> TargetT:
        ...


TypeMapperT = TypeVar("TypeMapperT", bound=TypeMapper[Any])


class TypeMapperMeta:
    __slots__ = "t"

    def __init__(self, t: Type[Any]) -> None:
        self.t = t


def type_mapper(
    map_for: type[Any], /, *, cache: Cache | None = None
) -> Callable[[type[TypeMapperT]], type[TypeMapperT]]:
    def decorator(cls: type[TypeMapperT]) -> type[TypeMapperT]:
        meta.set(cls, TypeMapperMeta(Type(map_for)))
        (cache or __user_cache__).add(TypeMapperMeta, cls)
        return cls

    return decorator


class DefaultTypeMapper(TypeMapper[object]):
    __slots__ = "runner"

    def __init__(self, runner: MappingRunner) -> None:
        self.runner = runner

    @override
    def map(
        self,
        src_expr: ExpressionNode,
        src_t: Type[SrcT],
        dest_expr: ExpressionNode,
        dest_t: Type[Any],
        src: SrcT,
        dest: Any | None,
        exc_grp: list[MappingError] | None,
    ) -> Any:
        if dest_t.cls in (list, tuple, set):
            return self._map_iterable(src_expr, src_t, dest_expr, dest_t, src, dest, exc_grp)
        if dest_t.cls is dict:
            return self._map_dict(src_expr, dest_expr, dest_t, src, dest, exc_grp)
        return self._map_object(src_expr, src_t, dest_expr, dest_t, src, dest, exc_grp)

    def _map_object(
        self,
        src_expr: ExpressionNode,
        src_t: Type[SrcT],
        dest_expr: ExpressionNode,
        dest_t: Type[Any],
        src: SrcT,
        dest: Any | None,
        exc_grp: list[MappingError] | None,
    ) -> Any:
        if dest_t.is_any:
            return self.runner.map(src_expr, src_t, dest_expr, src_t, src, None, exc_grp)
        if dest is None:
            try:
                dest = dest_t.new()
            except TypeError as e:
                exc = InstantiationError(dest_expr, dest_t)
                if exc_grp is None:
                    raise exc from e
                exc_grp.append(exc)
                return None
        sequence: MappingSequence[SrcT, object] | None = self.runner.sequences.get(
            MappingSequence.get_hash(src_t, dest_t), None
        )
        if sequence is not None:
            for func in sequence.head:
                func.func(src, dest)
        for dest_name, anno_t in dest_t.annotations().items():
            src_name = dest_name
            field_src_expr = self._get_expr(src_expr, src, src_name)
            field_dest_expr: ExpressionNode = getattr(dest_expr, dest_name)
            selected_t: Type[Any] | None = None
            src_value_expr = self._get_expr(ExpressionTree.new(src_t), src, src_name)
            if sequence is not None and dest_name in sequence.for_attrs:
                for_attr = sequence.for_attrs[dest_name]
                if isinstance(for_attr, IgnoreAttribute):
                    if not anno_t.nullable:
                        exc = IgnoreImpossibleError(field_dest_expr, anno_t)
                        if exc_grp is None:
                            raise exc
                        exc_grp.append(exc)
                        continue
                    setattr(dest, dest_name, None)
                    continue
                if isinstance(for_attr, MapFromAttribute):
                    src_value_expr = for_attr.src_expr
                    selected_t = for_attr.use_type
            if anno_t.is_union:
                if selected_t is None:
                    exc = UnionNotAllowedError(field_dest_expr, anno_t)
                    if exc_grp is None:
                        raise exc
                    exc_grp.append(exc)
                    continue
                if selected_t != anno_t and selected_t not in anno_t.union:
                    exc = TypeMismatchError(field_src_expr, field_dest_expr, selected_t, anno_t)
                    if exc_grp is None:
                        raise exc
                    exc_grp.append(exc)
                    continue
                anno_t = selected_t
            try:
                value = ExpressionTree.get_value(src_value_expr, src)
            except Exception as err:
                if not self._has_default_value(dest, dest_name):
                    if not anno_t.nullable:
                        exc = SourceNotFoundError(field_src_expr, field_dest_expr, anno_t)
                        if exc_grp is None:
                            raise exc from err
                        exc_grp.append(exc)
                        continue
                    else:
                        setattr(dest, dest_name, None)
                        continue
                else:
                    setattr(dest, dest_name, self._get_default_value(dest, dest_name))
                    continue
            setattr(
                dest,
                dest_name,
                self.runner.map(
                    field_src_expr,
                    Type(type(value)),
                    field_dest_expr,
                    anno_t,
                    value,
                    None,
                    exc_grp,
                ),
            )
        if sequence is not None:
            for func in sequence.tail:
                func.func(src, dest)
        return dest

    def _map_iterable(
        self,
        src_expr: ExpressionNode,
        src_t: Type[SrcT],
        dest_expr: ExpressionNode,
        dest_t: Type[list[DestT]] | Type[set[DestT]] | Type[tuple[DestT]],
        src: SrcT,
        dest: list[DestT] | set[DestT] | tuple[DestT] | None,
        exc_grp: list[MappingError] | None,
    ) -> list[DestT] | set[DestT] | tuple[DestT]:
        elems: list[DestT] = []
        if not self._is_iterable(src):
            exc = TypeNotIterableError(src_expr, dest_expr, src_t, dest_t)
            if exc_grp is None:
                raise exc
            exc_grp.append(exc)
            return None  # type: ignore
        for index, elem in enumerate(src):
            elems.append(
                self.runner.map(
                    src_expr[index],
                    Type.from_instance(elem),
                    dest_expr[index],
                    Type(dest_t.vars[0]),
                    elem,
                    None,
                    exc_grp,
                )
            )
        if dest is None:
            dest = dest_t.new(elems)
        elif isinstance(dest, list):
            dest.clear()
            dest.extend(elems)
        elif isinstance(dest, set):
            dest.clear()
            dest.update(elems)
        else:
            exc = ImmutableCollectionError(dest_expr)
            if exc_grp is None:
                raise exc
            exc_grp.append(exc)
            return None  # type: ignore
        return dest

    def _map_dict(
        self,
        src_expr: ExpressionNode,
        dest_expr: ExpressionNode,
        dest_t: Type[dict[str, DestT]],
        src: object,
        dest: dict[str, DestT] | None,
        exc_grp: list[MappingError] | None,
    ) -> dict[str, DestT]:
        if dest is None:
            dest = dest_t.new()
        for src_name, src_value in self._iter_obj(src):
            dest[src_name] = self.runner.map(
                src_expr[src_name],
                Type.from_instance(src_value),
                getattr(dest_expr, src_name),
                Type(dest_t.vars[1]),
                src_value,
                None,
                exc_grp,
            )
        return dest

    @staticmethod
    def _has_attr(obj: object, attr: str) -> bool:
        if isinstance(obj, dict):
            return attr in obj
        return hasattr(obj, attr)

    @staticmethod
    def _has_default_value(obj: object, attr: str) -> bool:
        return hasattr(type(obj), attr)

    @staticmethod
    def _get_default_value(obj: object, attr: str) -> Any:
        return getattr(type(obj), attr)

    @staticmethod
    def _get_value(obj: object, attr: str) -> Any:
        if isinstance(obj, dict):
            return obj[attr]  # pyright: ignore[reportUnknownVariableType]
        return getattr(obj, attr)

    @staticmethod
    def _get_expr(path: ExpressionNode, obj: Any, attr: str) -> ExpressionNode:
        if isinstance(obj, dict):
            return path[attr]
        return getattr(path, attr)

    @staticmethod
    def _iter_obj(obj: object) -> Iterable[tuple[str, Any]]:
        if isinstance(obj, dict):
            return obj.items()  # pyright: ignore[reportUnknownVariableType]
        return vars(obj).items()

    @staticmethod
    def _is_iterable(obj: Any) -> TypeGuard[Iterable[Any]]:
        return hasattr(obj, "__iter__")


class IntegerTypeMapper(TypeMapper[int]):
    __slots__ = "runner"

    def __init__(self, runner: MappingRunner) -> None:
        self.runner = runner

    @override
    def map(
        self,
        src_expr: ExpressionNode,
        src_t: Type[SrcT],
        dest_expr: ExpressionNode,
        dest_t: Type[int],
        src: SrcT,
        dest: int | None,
        exc_grp: list[MappingError] | None,
    ) -> int:
        try:
            return int(src)  # type: ignore
        except (ValueError, TypeError) as err:
            exc = ConvertionError(src_expr, dest_expr, src, Type(int))
            if exc_grp is None:
                raise exc from err
            exc_grp.append(exc)
            return None  # type: ignore


class FloatTypeMapper(TypeMapper[float]):
    __slots__ = "runner"

    def __init__(self, runner: MappingRunner) -> None:
        self.runner = runner

    @override
    def map(
        self,
        src_expr: ExpressionNode,
        src_t: Type[SrcT],
        dest_expr: ExpressionNode,
        dest_t: Type[float],
        src: SrcT,
        dest: float | None,
        exc_grp: list[MappingError] | None,
    ) -> float:
        try:
            return float(src)  # type: ignore
        except ValueError as err:
            exc = ConvertionError(src_expr, dest_expr, src, Type(float))
            if exc_grp is None:
                raise exc from err
            exc_grp.append(exc)
            return None  # type: ignore


class BoolTypeMapper(TypeMapper[bool]):
    __slots__ = "runner"

    def __init__(self, runner: MappingRunner) -> None:
        self.runner = runner

    @override
    def map(
        self,
        src_expr: ExpressionNode,
        src_t: Type[SrcT],
        dest_expr: ExpressionNode,
        dest_t: Type[bool],
        src: SrcT,
        dest: bool | None,
        exc_grp: list[MappingError] | None,
    ) -> bool:
        return bool(src)


class StringTypeMapper(TypeMapper[str]):
    __slots__ = "runner"

    def __init__(self, runner: MappingRunner) -> None:
        self.runner = runner

    @override
    def map(
        self,
        src_expr: ExpressionNode,
        src_t: Type[SrcT],
        dest_expr: ExpressionNode,
        dest_t: Type[str],
        src: SrcT,
        dest: str | None,
        exc_grp: list[MappingError] | None,
    ) -> str:
        return str(src)



from typing import Any, TypeVar

InstanceT = TypeVar("InstanceT")


class AttributeUtils:
    @staticmethod
    def get_cls_attrs(
        obj: type[Any], *, of_type: type[InstanceT] | tuple[type[InstanceT], ...] | None = None
    ) -> dict[str, InstanceT]:
        parent_attrs: dict[str, Any] = {}
        for parent in obj.__bases__:
            parent_attrs |= AttributeUtils.get_cls_attrs(parent, of_type=of_type)
        return parent_attrs | {
            name: attribute
            for name, attribute in vars(obj).items()
            if of_type is None or isinstance(attribute, of_type)
        }

    @staticmethod
    def get_instance_attrs(obj: object, *, of_type: type[InstanceT] | None = None) -> dict[str, InstanceT]:
        return {
            name: attribute
            for name, attribute in vars(obj).items()
            if of_type is None or isinstance(attribute, of_type)
        }



from graphlib import CycleError, TopologicalSorter
from typing import Any, Final, Protocol

from typing_extensions import override

from bolinette.core import Cache, Logger
from bolinette.core.command import Parser
from bolinette.core.environment import CoreSection, Environment, environment
from bolinette.core.exceptions import InitError
from bolinette.core.injection import Injection, injectable
from bolinette.core.injection.injection import InjectionEvent, injection_callback
from bolinette.core.injection.registration import InjectionStrategy
from bolinette.core.mapping import Mapper, type_mapper
from bolinette.core.mapping.mapper import BoolTypeMapper, FloatTypeMapper, IntegerTypeMapper, StringTypeMapper
from bolinette.core.types import Type


class _ExtensionModule(Protocol):
    __blnt_ext__: "Extension"


class Extension:
    def __init__(self, name: str, dependencies: "list[_ExtensionModule] | None" = None) -> None:
        self.name = name
        self.dependencies = [m.__blnt_ext__ for m in dependencies] if dependencies else []

    def add_cached(self, cache: Cache) -> None:
        pass

    @staticmethod
    def sort_extensions(extensions: "list[Extension]") -> "list[Extension]":
        sorter: TopologicalSorter[Extension] = TopologicalSorter()
        for ext in extensions:
            sorter.add(ext, *ext.dependencies)
        try:
            return list(sorter.static_order())
        except CycleError as e:
            raise InitError("A circular dependency was detected in the loaded extensions") from e


class _CoreExtension(Extension):
    def __init__(self) -> None:
        super().__init__("core")

    @override
    def add_cached(self, cache: Cache) -> None:
        environment("core", cache=cache)(CoreSection)

        injection_callback(cache=cache)(InjectionLogger)
        injectable(strategy="singleton", cache=cache)(Injection)
        injectable(strategy="singleton", cache=cache)(Parser)
        injectable(strategy="transient", match_all=True, cache=cache)(Logger)
        injectable(strategy="singleton", cache=cache)(Environment)

        injectable(strategy="singleton", cache=cache)(Mapper)
        type_mapper(int, cache=cache)(IntegerTypeMapper)
        type_mapper(float, cache=cache)(FloatTypeMapper)
        type_mapper(bool, cache=cache)(BoolTypeMapper)
        type_mapper(str, cache=cache)(StringTypeMapper)


core_ext: Final[Extension] = _CoreExtension()


class InjectionLogger:
    def __init__(self, logger: Logger[Injection]) -> None:
        self.logger = logger

    def __call__(self, event: InjectionEvent, type: Type[Any], strategy: InjectionStrategy, instance: Any) -> None:
        match event:
            case "instantiated":
                self.logger.debug(f"Instantiated {type} with strategy '{strategy}'")



from typing import Any

from typing_extensions import override


class BolinetteError(Exception):
    def __init__(self, message: str) -> None:
        Exception.__init__(self, message)
        self.message = message


class InitError(BolinetteError):
    def __init__(self, message: str) -> None:
        BolinetteError.__init__(self, message)
        self.message = message

    @override
    def __str__(self) -> str:
        return self.message


class InternalError(BolinetteError):
    pass


class ParameterError:
    def __init__(self, **params: str) -> None:
        self._error_params = params

    def _format_params(self, message: str, **values: Any) -> str:
        f_strings: list[str] = []
        for param, f_string in self._error_params.items():
            if param in values and values[param]:
                f_strings.append(f_string.replace("{}", str(values[param])))
        return ", ".join([*f_strings, message])


class TypingError(BolinetteError, ParameterError):
    def __init__(
        self,
        message: str,
        *,
        cls: str | None = None,
    ) -> None:
        ParameterError.__init__(self, cls="Type {}")
        BolinetteError.__init__(self, self._format_params(message, cls=cls))


class InjectionError(BolinetteError, ParameterError):
    def __init__(
        self,
        message: str,
        *,
        cls: type[Any] | None = None,
        func: Any | None = None,
        param: str | None = None,
    ) -> None:
        ParameterError.__init__(self, cls="Type {}", func="Callable {}", param="Parameter '{}'")
        BolinetteError.__init__(self, self._format_params(message, cls=cls, func=func, param=param))


class EnvironmentError(BolinetteError):
    pass


class InitMappingError(InitError):
    pass



from collections.abc import Callable
from typing import Any, Generic, Self, TypeVar

from bolinette.core import Cache, __user_cache__
from bolinette.core.expressions import ExpressionNode, ExpressionTree
from bolinette.core.mapping.sequence import (
    ForAttributeMapping,
    IgnoreAttribute,
    IncludeFromBase,
    MapFromAttribute,
    MappingFunction,
    MappingSequence,
)
from bolinette.core.types import Type

SrcT = TypeVar("SrcT", bound=object)
DestT = TypeVar("DestT", bound=object)


class MapFromOptions:
    def __init__(self, attr: MapFromAttribute) -> None:
        self.step = attr

    def use_type(self, cls: type[Any]) -> None:
        self.step.use_type = Type(cls)


class MappingOptions(Generic[SrcT, DestT]):
    def __init__(self, sequence: MappingSequence[SrcT, DestT], dest_expr: ExpressionNode) -> None:
        self.sequence = sequence
        self.dest_expr = dest_expr
        self.step: ForAttributeMapping | None = None

    def map_from(self, func: Callable[[SrcT], Any]) -> MapFromOptions:
        src_expr: ExpressionNode = func(ExpressionTree.new(self.sequence.src_t))  # type: ignore
        self.step = MapFromAttribute(src_expr, self.dest_expr)
        return MapFromOptions(self.step)

    def ignore(self) -> None:
        self.step = IgnoreAttribute(self.dest_expr)


class SequenceBuilder(Generic[SrcT, DestT]):
    def __init__(self, src: type[SrcT], dest: type[DestT]) -> None:
        self.sequence = MappingSequence(src, dest)

    def for_attr(
        self,
        func: Callable[[DestT], Any],
        options: Callable[[MappingOptions[SrcT, DestT]], MapFromOptions | None],
    ) -> Self:
        expr: ExpressionNode = func(ExpressionTree.new(self.sequence.dest_t))  # type: ignore
        opt: MappingOptions[SrcT, DestT] = MappingOptions(self.sequence, expr)
        options(opt)
        if opt.step is not None:
            self.sequence.add_for_attr(opt.step)
        return self

    def before_mapping(
        self,
        func: Callable[[SrcT, DestT], None],
    ) -> Self:
        step = MappingFunction(func)
        self.sequence.add_head_func(step)
        return self

    def after_mapping(
        self,
        func: Callable[[SrcT, DestT], None],
    ) -> Self:
        step = MappingFunction(func)
        self.sequence.add_tail_func(step)
        return self

    def include(self, src_cls: type[Any], dest_cls: type[Any]) -> Self:
        self.sequence.add_include(IncludeFromBase(Type(src_cls), Type(dest_cls)))
        return self


class Profile:
    def __init__(self) -> None:
        self._sequences: list[MappingSequence[Any, Any]] = []

    @property
    def sequences(self) -> list[MappingSequence[Any, Any]]:
        return [*self._sequences]

    def register(self, src: type[SrcT], dest: type[DestT]) -> SequenceBuilder[SrcT, DestT]:
        builder = SequenceBuilder(src, dest)
        self._sequences.append(builder.sequence)
        return builder


ProfileT = TypeVar("ProfileT", bound=Profile)


def mapping(*, cache: Cache | None = None) -> Callable[[type[ProfileT]], type[ProfileT]]:
    def decorator(cls: type[ProfileT]) -> type[ProfileT]:
        (cache or __user_cache__).add(Profile, cls)
        return cls

    return decorator



from collections.abc import Callable
from typing import Any, Generic, TypeVar, get_args, get_origin

from bolinette.core import Cache, meta
from bolinette.core.expressions import ExpressionNode, ExpressionTree
from bolinette.core.injection import Injection
from bolinette.core.injection.context import InjectionContext

MockedT = TypeVar("MockedT")
SetupT = TypeVar("SetupT")


class _MockedMeta(Generic[MockedT]):
    def __init__(self, cls: type[MockedT]) -> None:
        self.cls = cls
        self.dummy = False
        self._attrs: dict[str, Any] = {}

    def _get_dummy(self, key: str) -> None | Callable[..., None]:
        if hasattr(self.cls, key) and callable(getattr(self.cls, key)):
            return lambda *args, **kwargs: None
        return None

    def __contains__(self, key: str) -> bool:
        return key in self._attrs or self.dummy

    def __getitem__(self, key: str) -> Any:
        if key not in self._attrs and self.dummy:
            return self._get_dummy(key)
        return self._attrs[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._attrs[key] = value


class _MockWrapper(Generic[MockedT]):
    def __init__(
        self,
        cls: type[MockedT],
    ) -> None:
        self._cls = self._setup_mocked_cls(cls)
        self.instance = self._cls()

    @staticmethod
    def _get_mocked_attr(_cls: type[MockedT], instance: MockedT, name: str) -> Any:
        if name == "__class__":
            return _cls
        _meta: _MockedMeta[Any] = meta.get(type(instance), _MockedMeta)
        if name in _meta:
            return _meta[name]
        raise KeyError(f"'{name}' attribute has not been mocked in {_cls}")

    @staticmethod
    def _setup_mocked_cls(_cls: type[MockedT]) -> type[MockedT]:
        def _get_attr(instance: object, name: str) -> Any:
            return _MockWrapper._get_mocked_attr(_cls, instance, name)

        _t = type(f"{_cls.__name__}__Mocked", (_cls,), {})
        _t.__init__ = lambda _: None  # type: ignore
        _t.__repr__ = lambda _: f"<Mocked[{_cls.__name__}]>"  # type: ignore
        _t.__getattribute__ = _get_attr
        meta.set(_t, _MockedMeta(_cls))
        return _t  # type: ignore

    def setup(self, func: Callable[[MockedT], SetupT], value: SetupT) -> "_MockWrapper[MockedT]":
        expr: ExpressionNode = func(ExpressionTree.new())  # type: ignore
        ExpressionTree.ensure_attribute_chain(expr)
        name = ExpressionTree.get_attribute(expr)
        _meta: _MockedMeta[Any] = meta.get(self._cls, _MockedMeta)
        _meta[name] = value
        return self

    def dummy(self, value: bool = True) -> "_MockWrapper[MockedT]":
        _meta: _MockedMeta[Any] = meta.get(self._cls, _MockedMeta)
        _meta.dummy = value
        return self


class Mock:
    def __init__(self, *, inject: Injection | None = None, cache: Cache | None = None) -> None:
        self._inject = inject or Injection(cache or Cache(), InjectionContext())
        self._mocked: dict[type[Any], _MockWrapper[Any]] = {}

    @staticmethod
    def _get_generic_params(
        _cls: type[MockedT],
    ) -> tuple[type[MockedT], tuple[Any, ...]]:
        if origin := get_origin(_cls):
            params: tuple[Any, ...] = ()
            for arg in get_args(_cls):
                params = (*params, arg)
            return origin, params
        return _cls, ()

    def mock(self, cls: type[MockedT], *, match_all: bool = False) -> _MockWrapper[MockedT]:
        origin, _ = self._get_generic_params(cls)
        if origin in self._mocked:
            mocked = self._mocked[origin]
        else:
            mocked = _MockWrapper(origin)
            self._mocked[origin] = mocked
            self._inject.add(cls, "singleton", instance=mocked.instance, match_all=match_all)
        return mocked

    @property
    def injection(self) -> Injection:
        return self._inject



from collections.abc import Callable
from typing import Protocol, TypeVar

from bolinette.core import Cache, __user_cache__, meta


class EnvSectionMeta:
    def __init__(self, name: str) -> None:
        self.name = name


class EnvironmentSection(Protocol):
    def __init__(self) -> None:
        pass


EnvT = TypeVar("EnvT", bound=EnvironmentSection)


def environment(name: str, *, cache: Cache | None = None) -> Callable[[type[EnvT]], type[EnvT]]:
    def decorator(cls: type[EnvT]) -> type[EnvT]:
        meta.set(cls, EnvSectionMeta(name))
        (cache or __user_cache__).add(EnvironmentSection, cls)
        return cls

    return decorator


class CoreSection:
    debug: bool = False



from bolinette.core.injection.decorators import init_method as init_method, require as require, injectable as injectable
from bolinette.core.injection.injection import Injection as Injection
from bolinette.core.injection.resolver import injection_arg_resolver as injection_arg_resolver



import sys
from datetime import datetime
from typing import Generic, Protocol, TypeVar

from bolinette.core import Cache, GenericMeta, meta
from bolinette.core.injection import init_method

T_Contra = TypeVar("T_Contra", contravariant=True)
T = TypeVar("T")


class SupportsWrite(Protocol[T_Contra]):
    def write(self, __s: T_Contra) -> object:
        pass


class ConsoleColorCode:
    Reset = "\x1b[0m"
    Bright = "\x1b[1m"
    Dim = "\x1b[2m"
    Underscore = "\x1b[4m"
    Blink = "\x1b[5m"
    Reverse = "\x1b[7m"
    Hidden = "\x1b[8m"

    FgBlack = "\x1b[30m"
    FgRed = "\x1b[31m"
    FgGreen = "\x1b[32m"
    FgYellow = "\x1b[33m"
    FgBlue = "\x1b[34m"
    FgMagenta = "\x1b[35m"
    FgCyan = "\x1b[36m"
    FgWhite = "\x1b[37m"

    BgBlack = "\x1b[40m"
    BgRed = "\x1b[41m"
    BgGreen = "\x1b[42m"
    BgYellow = "\x1b[43m"
    BgBlue = "\x1b[44m"
    BgMagenta = "\x1b[45m"
    BgCyan = "\x1b[46m"
    BgWhite = "\x1b[47m"


class Logger(Generic[T]):
    def __init__(self, cache: Cache | None = None) -> None:
        self._cache = cache or Cache()
        self._package = "<Logger>"

    @init_method
    def _init(self) -> None:
        if gen_meta := meta.get(self, GenericMeta, default=None):
            if len(gen_meta):
                self._package = gen_meta[0].__name__

    def _log(
            self,
            prefix: str,
            text: str,
            color: str | None = None,
            file: SupportsWrite[str] | None = None,
    ):
        strings: list[str] = [f"{datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%f')}",
                              f"{ConsoleColorCode.Bright}{color}{prefix}{ConsoleColorCode.Reset}",
                              f"[{ConsoleColorCode.FgGreen}{self._package}{ConsoleColorCode.Reset}]", text]
        print(*strings, file=file)

    def warning(self, *values: str, sep: str | None = None) -> None:
        self._log("WARN", (sep or " ").join(values), color=ConsoleColorCode.FgYellow)

    def info(self, *values: str, sep: str | None = None) -> None:
        self._log("INFO", (sep or " ").join(values), color=ConsoleColorCode.FgGreen)

    def debug(self, *values: str, sep: str | None = None) -> None:
        if self._cache.debug:
            self._log(
                "DEBUG",
                (sep or " ").join(values),
                color=ConsoleColorCode.FgBlue,
            )

    def error(self, *values: str, sep: str | None = None) -> None:
        self._log(
            "ERROR",
            (sep or " ").join(values),
            color=ConsoleColorCode.FgRed,
            file=sys.stderr,
        )



from bolinette.core import expressions
from bolinette.core.exceptions import BolinetteError, ParameterError


class ExpressionError(BolinetteError, ParameterError):
    def __init__(self, expr: "expressions.ExpressionNode", message: str) -> None:
        ParameterError.__init__(self, expr="Expression {}")
        BolinetteError.__init__(self, self._format_params(message, expr=expr))


class MaxDepthExpressionError(ExpressionError):
    def __init__(self, expr: "expressions.ExpressionNode") -> None:
        super().__init__(expr, "Expression exceeds allowed depth")


class AttributeChainError(ExpressionError):
    def __init__(self, expr: "expressions.ExpressionNode") -> None:
        super().__init__(expr, "Expression is excepted to be a chain of attribute access")
